""" Genes """
import sys
from typing import List
from functools import lru_cache

import numpy as np

from AnyQt.QtGui import QFont, QColor
from AnyQt.QtCore import Qt, QSize, QTimer, QVariant, QModelIndex, QAbstractTableModel
from AnyQt.QtWidgets import QStyle, QSplitter, QTableView, QHeaderView, QAbstractItemView

from Orange.data import Table, Domain, StringVariable, DiscreteVariable
from Orange.data import filter as table_filter
from Orange.widgets.gui import (
    LinkRole,
    LinkStyledItemDelegate,
    vBox,
    rubber,
    checkBox,
    comboBox,
    lineEdit,
    widgetBox,
    auto_commit,
    widgetLabel,
)
from Orange.widgets.utils import itemmodels
from Orange.widgets.widget import OWWidget
from Orange.widgets.settings import Setting, ContextSetting, DomainContextHandler
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.utils.concurrent import TaskState, ConcurrentWidgetMixin

from orangecontrib.bioinformatics.ncbi import taxonomy
from orangecontrib.bioinformatics.ncbi.gene import ENTREZ_ID, Gene, GeneMatcher
from orangecontrib.bioinformatics.widgets.utils.data import (
    TAX_ID,
    GENE_ID_COLUMN,
    GENE_ID_ATTRIBUTE,
    GENE_AS_ATTRIBUTE_NAME,
)

NCBI_DETAIL_LINK = 'http://www.ncbi.nlm.nih.gov/sites/entrez?Db=gene&Cmd=ShowDetailView&TermToSearch={}'
HEADER = [
    ['Input ID', 'Entrez ID', 'Name', 'Description', 'Synonyms', 'Other IDs'],
    ['input_identifier', 'gene_id', 'symbol', 'description', 'synonyms', 'db_refs'],
]


def run_gene_matcher(gene_matcher: GeneMatcher, state: TaskState):
    current_iter = 0
    max_iter = len(gene_matcher.genes)

    def callback():
        nonlocal current_iter
        current_iter += 1
        state.set_progress_value(100 * (current_iter / max_iter))

    state.set_status("Working ...")
    gene_matcher._progress_callback = callback
    gene_matcher.match_genes()


class GeneInfoModel(itemmodels.PyTableModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.header_labels, self.gene_attributes = HEADER
        self.setHorizontalHeaderLabels(self.header_labels)

        try:
            # note: make sure gene_id is set in owgenes_header
            self.entrez_column_index = self.gene_attributes.index('gene_id')
        except ValueError:
            raise ValueError("Make sure 'gene_id' is set in header")

        self.genes = None
        self.table = None

        self.font = QFont()
        self.font.setUnderline(True)
        self.color = QColor(Qt.blue)

        @lru_cache(maxsize=10000)
        def _row_instance(row, column):
            return self[int(row)][int(column)]

        self._row_instance = _row_instance

    def initialize(self, list_of_genes):
        self.genes = list_of_genes
        self.__table_from_genes([gene for gene in list_of_genes if gene.gene_id])
        self.update_model()

    def rowCount(self, parent=QModelIndex()):
        return 0 if parent.isValid() else len(self._table)

    def columnCount(self, parent=QModelIndex()):
        return 0 if (parent.isValid() or self._table.size == 0) else self._table.shape[1]

    def clear(self):
        self.beginResetModel()
        self._table = np.array([[]])
        self.resetSorting()
        self._roleData.clear()
        self.endResetModel()

    def data(
        self,
        index,
        role,
        _str=str,
        _Qt_DisplayRole=Qt.DisplayRole,  # noqa: N803
        _Qt_EditRole=Qt.EditRole,
        _Qt_FontRole=Qt.FontRole,
        _Qt_ForegroundRole=Qt.ForegroundRole,
        _LinkRolee=LinkRole,
        _recognizedRoles=frozenset([Qt.DisplayRole, Qt.EditRole, Qt.FontRole, Qt.ForegroundRole, LinkRole]),
    ):

        if role not in _recognizedRoles:
            return None

        row, col = index.row(), index.column()
        if not 0 <= row <= self.rowCount():
            return None
        row = self.mapToSourceRows(row)

        try:
            # value = self[row][col]
            value = self._row_instance(row, col)
        except IndexError:
            return

        if role == Qt.DisplayRole:
            return QVariant(str(value))
        elif role == Qt.ToolTipRole:
            return QVariant(str(value))

        if col == self.entrez_column_index:
            if role == _Qt_ForegroundRole:
                return self.color
            elif role == _Qt_FontRole:
                return self.font
            elif role == _LinkRolee:
                return NCBI_DETAIL_LINK.format(value)

    def __table_from_genes(self, list_of_genes: List[Gene]) -> None:
        def to_list(gene: Gene) -> List[str]:
            _, header_tags = HEADER

            def parse_attribute(tag):
                gene_attr = getattr(gene, '{}'.format(tag))

                if isinstance(gene_attr, dict):
                    # note: db_refs are stored as dicts
                    gene_attr = (
                        ', '.join('{}: {}'.format(key, val) for (key, val) in gene_attr.items()) if gene_attr else ' '
                    )
                elif isinstance(gene_attr, list):
                    # note: synonyms are stored as lists
                    gene_attr = ', '.join(gene_attr) if gene_attr else ' '

                return gene_attr

            return [parse_attribute(tag) for tag in header_tags]

        self.table = np.asarray([to_list(gene) for gene in list_of_genes])

    def get_filtered_genes(self):
        return list(self._table[:, self.entrez_column_index]) if self._table.size else []

    def filter_table(self, filter_pattern: str):
        selection = np.full(self.table.shape, True)
        for search_word in filter_pattern.split():
            match_result = np.core.defchararray.find(np.char.lower(self.table), search_word.lower()) >= 0
            selection = selection & match_result
        return selection

    def update_model(self, filter_pattern=''):
        # clear cache if model changes
        self._row_instance.cache_clear()

        if self.table.size:
            self.wrap(self.table[self.filter_table(filter_pattern).any(axis=1), :])


class UnknownGeneInfoModel(itemmodels.PyListModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.header_labels = ['IDs from the input data without corresponding Entrez ID']
        self.genes = []

    def initialize(self, list_of_genes):
        self.genes = list_of_genes
        self.wrap([', '.join([gene.input_identifier for gene in list_of_genes if not gene.gene_id])])

    def data(self, index, role=Qt.DisplayRole):
        row = index.row()
        if role in [self.list_item_role, Qt.EditRole] and self._is_index_valid(index):
            return self[row]
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignLeft | Qt.AlignTop
        elif self._is_index_valid(row):
            return self._other_data[row].get(role, None)

    def headerData(self, section, orientation, role=Qt.DisplayRole):

        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.header_labels[section]
        return QAbstractTableModel.headerData(self, section, orientation, role)


class OWGenes(OWWidget, ConcurrentWidgetMixin):
    name = "Genes"
    description = "Tool for working with genes"
    icon = "../widgets/icons/OWGeneInfo.svg"
    priority = 40
    want_main_area = True

    selected_organism: int = Setting(11)
    search_pattern: str = Setting('')
    exclude_unmatched = Setting(True)
    replace_id_with_symbol = Setting(True)
    auto_commit = Setting(True)

    settingsHandler = DomainContextHandler()
    selected_gene_col = ContextSetting(None)
    use_attr_names = ContextSetting(True)

    replaces = ['orangecontrib.bioinformatics.widgets.OWGeneNameMatcher.OWGeneNameMatcher']

    class Inputs:
        data_table = Input("Data", Table)

    class Outputs:
        data_table = Output("Data", Table)
        gene_matcher_results = Output("Genes", Table)

    class Information(OWWidget.Information):
        pass

    def sizeHint(self):
        return QSize(1280, 960)

    def __init__(self):
        OWWidget.__init__(self)
        ConcurrentWidgetMixin.__init__(self)

        # ATTRIBUTES #
        self.target_database = ENTREZ_ID

        # input data
        self.input_data = None
        self.input_genes = None
        self.tax_id = None
        self.column_candidates = []

        # input options
        self.organisms = []

        # gene matcher
        self.gene_matcher = None

        # progress bar
        self.progress_bar = None

        self._timer = QTimer()
        self._timer.timeout.connect(self._apply_filter)
        self._timer.setSingleShot(True)

        # GUI SECTION #

        # Control area
        self.info_box = widgetLabel(widgetBox(self.controlArea, "Info", addSpace=True), 'No data on input.\n')

        organism_box = vBox(self.controlArea, 'Organism')
        self.organism_select_combobox = comboBox(
            organism_box, self, 'selected_organism', callback=self.on_input_option_change
        )

        self.get_available_organisms()
        self.organism_select_combobox.setCurrentIndex(self.selected_organism)

        box = widgetBox(self.controlArea, 'Gene IDs in the input data')
        self.gene_columns_model = itemmodels.DomainModel(valid_types=(StringVariable, DiscreteVariable))
        self.gene_column_combobox = comboBox(
            box,
            self,
            'selected_gene_col',
            label='Stored in data column',
            model=self.gene_columns_model,
            sendSelectedValue=True,
            callback=self.on_input_option_change,
        )

        self.attr_names_checkbox = checkBox(
            box,
            self,
            'use_attr_names',
            'Stored as feature (column) names',
            disables=[(-1, self.gene_column_combobox)],
            callback=self.on_input_option_change,
        )

        self.gene_column_combobox.setDisabled(bool(self.use_attr_names))

        output_box = vBox(self.controlArea, 'Output')

        # separator(output_box)
        # output_box.layout().addWidget(horizontal_line())
        # separator(output_box)
        self.exclude_radio = checkBox(
            output_box, self, 'exclude_unmatched', 'Exclude unmatched genes', callback=self.commit
        )

        self.replace_radio = checkBox(
            output_box, self, 'replace_id_with_symbol', 'Replace feature IDs with gene names', callback=self.commit
        )

        auto_commit(self.controlArea, self, "auto_commit", "&Commit", box=False)

        rubber(self.controlArea)

        # Main area
        self.filter = lineEdit(
            self.mainArea, self, 'search_pattern', 'Filter:', callbackOnType=True, callback=self.handle_filter_callback
        )
        # rubber(self.radio_group)
        self.mainArea.layout().addWidget(self.filter)

        # set splitter
        self.splitter = QSplitter()
        self.splitter.setOrientation(Qt.Vertical)

        self.table_model = GeneInfoModel()
        self.table_view = QTableView()
        self.table_view.setAlternatingRowColors(True)
        self.table_view.viewport().setMouseTracking(True)
        self.table_view.setSortingEnabled(True)
        self.table_view.setShowGrid(False)
        self.table_view.verticalHeader().hide()
        # self.table_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.unknown_model = UnknownGeneInfoModel()

        self.unknown_view = QTableView()
        self.unknown_view.setModel(self.unknown_model)
        self.unknown_view.verticalHeader().hide()
        self.unknown_view.setShowGrid(False)
        self.unknown_view.setSelectionMode(QAbstractItemView.NoSelection)
        self.unknown_view.horizontalHeader().setSectionResizeMode(QHeaderView.Stretch)

        self.splitter.addWidget(self.table_view)
        self.splitter.addWidget(self.unknown_view)

        self.splitter.setStretchFactor(0, 90)
        self.splitter.setStretchFactor(1, 10)

        self.mainArea.layout().addWidget(self.splitter)

    def handle_filter_callback(self):
        self._timer.stop()
        self._timer.start(500)

    def _apply_filter(self):
        # filter only if input data is present and model is populated
        if self.table_model.table is not None:
            self.table_model.update_model(filter_pattern=str(self.search_pattern))
            self.commit()

    def __reset_widget_state(self):
        self.table_view.clearSpans()
        self.table_view.setModel(None)
        self.table_model.clear()
        self.unknown_model.clear()
        self._update_info_box()

    def _update_info_box(self):

        if self.input_genes and self.gene_matcher:
            num_genes = len(self.gene_matcher.genes)
            known_genes = len(self.gene_matcher.get_known_genes())

            info_text = (
                '{} genes in input data\n'
                '{} genes match Entrez database\n'
                '{} genes with match conflicts\n'.format(num_genes, known_genes, num_genes - known_genes)
            )

        else:
            info_text = 'No data on input.'

        self.info_box.setText(info_text)

    def on_done(self, _):
        # update info box
        self._update_info_box()

        # set output options
        self.toggle_radio_options()

        # set known genes
        self.table_model.initialize(self.gene_matcher.genes)
        self.table_view.setModel(self.table_model)
        self.table_view.selectionModel().selectionChanged.connect(self.commit)
        self.table_view.setSelectionBehavior(QAbstractItemView.SelectRows)

        self.table_view.setItemDelegateForColumn(
            self.table_model.entrez_column_index, LinkStyledItemDelegate(self.table_view)
        )
        v_header = self.table_view.verticalHeader()
        option = self.table_view.viewOptions()
        size = self.table_view.style().sizeFromContents(QStyle.CT_ItemViewItem, option, QSize(20, 20), self.table_view)

        v_header.setDefaultSectionSize(size.height() + 2)
        v_header.setMinimumSectionSize(5)
        self.table_view.horizontalHeader().setStretchLastSection(True)

        # set unknown genes
        self.unknown_model.initialize(self.gene_matcher.genes)
        self.unknown_view.verticalHeader().setStretchLastSection(True)

        self._apply_filter()

    def get_available_organisms(self):
        available_organism = sorted(
            ((tax_id, taxonomy.name(tax_id)) for tax_id in taxonomy.common_taxids()), key=lambda x: x[1]
        )

        self.organisms = [tax_id[0] for tax_id in available_organism]
        self.organism_select_combobox.addItems([tax_id[1] for tax_id in available_organism])

    def gene_names_from_table(self):
        """ Extract and return gene names from `Orange.data.Table`.
        """
        self.input_genes = []
        if self.input_data:
            if self.use_attr_names:
                self.input_genes = [str(attr.name).strip() for attr in self.input_data.domain.attributes]
            else:
                if self.selected_gene_col is None:
                    self.selected_gene_col = self.gene_column_identifier()

                self.input_genes = [
                    str(e[self.selected_gene_col]) for e in self.input_data if not np.isnan(e[self.selected_gene_col])
                ]

    def _update_gene_matcher(self):
        self.gene_names_from_table()

        self.gene_matcher = GeneMatcher(self.get_selected_organism(), auto_start=False)
        self.gene_matcher.genes = self.input_genes
        # self.gene_matcher.organism = self.get_selected_organism()

    def get_selected_organism(self):
        return self.organisms[self.selected_organism]

    def _run(self):
        if self.gene_matcher is not None:
            self.start(run_gene_matcher, self.gene_matcher)

    def on_input_option_change(self):
        self.__reset_widget_state()
        self._update_gene_matcher()
        self._run()

    def gene_column_identifier(self):
        """
        Get most suitable column that stores genes. If there are
        several suitable columns, select the one with most unique
        values. Take the best one.
        """

        # candidates -> (variable, num of unique values)
        candidates = (
            (col, np.unique(self.input_data.get_column_view(col)[0]).size)
            for col in self.gene_columns_model
            if isinstance(col, DiscreteVariable) or isinstance(col, StringVariable)
        )

        best_candidate, _ = sorted(candidates, key=lambda x: x[1])[-1]
        return best_candidate

    def find_genes_location(self):
        """ Try locate the genes in the input data when we first load the data.

            Proposed rules:
                - when no suitable feature names are present, check the columns.
                - find the most suitable column, that is, the one with most unique values.

        """
        domain = self.input_data.domain
        if not domain.attributes:
            if self.selected_gene_col is None:
                self.selected_gene_col = self.gene_column_identifier()
                self.use_attr_names = False

    @Inputs.data_table
    def handle_input(self, data):
        self.closeContext()
        self.input_data = None
        self.input_genes = None
        self.__reset_widget_state()
        self.gene_columns_model.set_domain(None)
        self.selected_gene_col = None

        if data:
            self.input_data = data
            self.gene_columns_model.set_domain(self.input_data.domain)

            # check if input table has tax_id, human is used if tax_id is not found
            self.tax_id = str(self.input_data.attributes.get(TAX_ID, '9606'))
            # check for gene location. Default is that genes are attributes in the input table.
            self.use_attr_names = self.input_data.attributes.get(GENE_AS_ATTRIBUTE_NAME, self.use_attr_names)

            if self.tax_id in self.organisms and not self.selected_organism:
                self.selected_organism = self.organisms.index(self.tax_id)

            self.openContext(self.input_data.domain)
            self.find_genes_location()
            self.on_input_option_change()

    def commit(self):
        selection = self.table_view.selectionModel().selectedRows(self.table_model.entrez_column_index)

        selected_genes = [row.data() for row in selection]
        if not len(selected_genes):
            selected_genes = self.table_model.get_filtered_genes()

        gene_ids = self.get_target_ids()
        known_genes = [gid for gid in gene_ids if gid != '?']

        table = None
        gm_table = None
        if known_genes:
            # Genes are in rows (we have a column with genes).
            if not self.use_attr_names:

                if self.target_database in self.input_data.domain:
                    gene_var = self.input_data.domain[self.target_database]
                    metas = self.input_data.domain.metas
                else:
                    gene_var = StringVariable(self.target_database)
                    metas = self.input_data.domain.metas + (gene_var,)

                domain = Domain(self.input_data.domain.attributes, self.input_data.domain.class_vars, metas)

                table = self.input_data.transform(domain)
                col, _ = table.get_column_view(gene_var)
                col[:] = gene_ids

                # filter selected rows
                selected_genes_set = set(selected_genes)
                selected_rows = [
                    row_index for row_index, row in enumerate(table) if str(row[gene_var]) in selected_genes_set
                ]

                # handle table attributes
                table.attributes[TAX_ID] = self.get_selected_organism()
                table.attributes[GENE_AS_ATTRIBUTE_NAME] = False
                table.attributes[GENE_ID_COLUMN] = self.target_database
                table = table[selected_rows] if selected_rows else table

                if self.exclude_unmatched:
                    # create filter from selected column for genes
                    only_known = table_filter.FilterStringList(gene_var, known_genes)
                    # apply filter to the data
                    table = table_filter.Values([only_known])(table)

                self.Outputs.data_table.send(table)

            # genes are are in columns (genes are features).
            else:
                domain = self.input_data.domain.copy()
                table = self.input_data.transform(domain)

                for gene in self.gene_matcher.genes:
                    if gene.input_identifier in table.domain:

                        table.domain[gene.input_identifier].attributes[self.target_database] = (
                            str(gene.gene_id) if gene.gene_id else '?'
                        )

                        if self.replace_id_with_symbol:
                            try:
                                table.domain[gene.input_identifier].name = str(gene.symbol)
                            except AttributeError:
                                # TODO: missing gene symbol, need to handle this?
                                pass

                # filter selected columns
                selected_genes_set = set(selected_genes)
                selected = [
                    column
                    for column in table.domain.attributes
                    if self.target_database in column.attributes
                    and str(column.attributes[self.target_database]) in selected_genes_set
                ]

                output_attrs = table.domain.attributes

                if selected:
                    output_attrs = selected

                if self.exclude_unmatched:
                    known_genes_set = set(known_genes)
                    output_attrs = [
                        col for col in output_attrs if col.attributes[self.target_database] in known_genes_set
                    ]

                domain = Domain(output_attrs, table.domain.class_vars, table.domain.metas)

                table = table.from_table(domain, table)

                # handle table attributes
                table.attributes[TAX_ID] = self.get_selected_organism()
                table.attributes[GENE_AS_ATTRIBUTE_NAME] = True
                table.attributes[GENE_ID_ATTRIBUTE] = self.target_database

            gm_table = self.gene_matcher.to_data_table(selected_genes=selected_genes if selected_genes else None)

        self.Outputs.data_table.send(table)
        self.Outputs.gene_matcher_results.send(gm_table)

    def toggle_radio_options(self):
        self.replace_radio.setEnabled(bool(self.use_attr_names))

        if self.gene_matcher.genes:
            # enable checkbox if unknown genes are detected
            self.exclude_radio.setEnabled(len(self.gene_matcher.genes) != len(self.gene_matcher.get_known_genes()))
            self.exclude_unmatched = len(self.gene_matcher.genes) != len(self.gene_matcher.get_known_genes())

    def get_target_ids(self):
        return [str(gene.gene_id) if gene.gene_id else '?' for gene in self.gene_matcher.genes]


if __name__ == "__main__":

    def main_test():
        from AnyQt.QtWidgets import QApplication
        import sys

        app = QApplication([])
        w = OWGenes()
        if len(sys.argv) > 1:
            data = Table(sys.argv[1])
            w.handle_input(data)
        w.show()
        w.raise_()
        r = app.exec_()
        w.saveSettings()
        return r

    sys.exit(main_test())



""" Resolwe module """
from .utils import GENAPI_CACHE, CACHE_BACKEND, RESOLWEAPI_CACHE
from .genapi import DEFAULT_EMAIL, DEFAULT_PASSWD, GenAPI
from .resapi import ResolweAPI

__all__ = ('DEFAULT_EMAIL', 'DEFAULT_PASSWD')


class ResolweAuthException(Exception):
    """ A login error occurred. """


class ResolweServerTypeException(Exception):
    """ Unknown server type """


class ResolweDataObjectsNotFound(Exception):
    """ Data Objects not found """


def connect(username=None, password=None, url=None, server_type='resolwe'):
    """Connect to Resolwe server

    :param username:
    :type username: str

    :param password:
    :type password: str

    :param url:
    :type url: str

    :param server_type: genesis or resolwe
    :type server_type: str

    :return: Instance of GenAPI or ResolweAPI
    """

    if server_type == 'resolwe':
        _api = ResolweAPI
    elif server_type == 'genesis':
        _api = GenAPI
    else:
        raise ResolweServerTypeException(f'Unknown server type {server_type}')

    try:
        return _api(username, password, url)
    except Exception as e:
        raise ResolweAuthException(e.args[0]) from e



import unittest

from orangewidget.tests.base import WidgetTest

from AnyQt.QtCore import Qt, QPoint
from AnyQt.QtTest import QTest, QSignalSpy
from AnyQt.QtWidgets import QTreeWidget, QTreeWidgetItem, QTreeWidgetItemIterator

from Orange.data import Table, Domain, StringVariable, DiscreteVariable
from Orange.widgets.widget import Msg, OWWidget
from Orange.widgets.settings import SettingProvider
from Orange.widgets.tests.utils import simulate

from orangecontrib.bioinformatics.widgets.components import GeneSetSelection
from orangecontrib.bioinformatics.widgets.utils.data import TableAnnotation


class MockWidget(OWWidget):
    name = 'Mock'
    want_main_area = False
    selection_component = SettingProvider(GeneSetSelection)

    class Error(OWWidget.Error):
        custom_gene_sets_table_format = Msg('FooBar')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.selection_component = GeneSetSelection(self, self.controlArea)


class TestGeneSetSelection(WidgetTest):
    def setUp(self):
        self.widget = MockWidget()
        self.component = self.widget.selection_component

        self.mouse_tax_id: str = '10090'
        self.human_tax_id: str = '9606'
        self.dicty_tax_id: str = '44689'

        self.default_selection = [('GO', 'biological_process')]
        self.marker_genes = [('Marker Genes', 'Panglao'), ('Marker Genes', 'CellMarker')]
        self.gene_ontologies = [
            ('GO', 'cellular_component'),
            ('GO', 'biological_process'),
            ('GO', 'molecular_function'),
        ]

    def __item_iterator(self):
        iterator = QTreeWidgetItemIterator(self.component.hierarchy_tree_widget, QTreeWidgetItemIterator.All)

        while iterator.value():
            item: QTreeWidgetItem = iterator.value()
            if not item.childCount():
                yield item
            iterator += 1

    def __initialization(self, tax_id: str) -> None:
        self.component.initialize(tax_id)

        # test if gene sets are shown
        gene_set_items = {item.hierarchy for item in self.__item_iterator()}
        self.assertTrue(len(gene_set_items))
        self.assertEqual(self.component.gene_sets.hierarchies(), gene_set_items)

        # we expect gene ontologies
        for go in self.gene_ontologies:
            self.assertIn(go, gene_set_items)

        # test default selection
        selection = self.component._get_selection()
        self.assertTrue(len(selection))
        self.assertEqual(selection, self.default_selection)

    def test_dictyostelium(self):
        self.__initialization(self.dicty_tax_id)
        self.assertEqual(self.component.gene_sets.common_org(), self.dicty_tax_id)

    def test_mus_musculus(self):
        self.__initialization(self.mouse_tax_id)
        self.assertEqual(self.component.gene_sets.common_org(), self.mouse_tax_id)

    def test_homo_sapiens(self):
        self.__initialization(self.human_tax_id)
        self.assertEqual(self.component.gene_sets.common_org(), self.human_tax_id)

    def test_selection(self):
        self.__initialization(self.human_tax_id)

        panglao, _ = self.marker_genes

        new_selection = None
        for item in self.__item_iterator():
            if item.hierarchy == panglao:
                new_selection = item

        tree_widget: QTreeWidget = self.component.hierarchy_tree_widget
        # obtain the rectangular coordinates
        rect = tree_widget.visualItemRect(new_selection)
        emitted_signals: QSignalSpy = QSignalSpy(self.component.selection_changed)

        # simulate click on item but not in checkbox area.
        # checkbox status is not changed and signal is not emitted
        click_item_text = QPoint(rect.x() + 100, rect.y() + 5)
        QTest.mouseClick(tree_widget.viewport(), Qt.LeftButton, Qt.NoModifier, click_item_text)
        self.assertFalse(len(emitted_signals))
        self.assertEqual(self.component._get_selection(), self.default_selection)

        # simulate click on items checkbox.
        # checkbox status is changed and signal is emitted
        click_checkbox = QPoint(rect.x() + 10, rect.y() + 5)
        QTest.mouseClick(tree_widget.viewport(), Qt.LeftButton, Qt.NoModifier, click_checkbox)
        self.assertTrue(len(emitted_signals))
        self.assertNotEqual(self.component._get_selection(), self.default_selection)
        self.assertEqual(set(self.component._get_selection()), set(self.default_selection + [new_selection.hierarchy]))

    def test_custom_gene_sets(self):
        domain = Domain(
            [], metas=[StringVariable('genes')], class_vars=DiscreteVariable('cell_type', values=['foo', 'bar'])
        )
        data = Table.from_list(domain, [[0, 'gene1'], [1, 'gene2'], [1, 'gene3']])
        data.name = 'custom_gene_sets'
        data.attributes[TableAnnotation.tax_id] = self.human_tax_id
        data.attributes[TableAnnotation.gene_as_attr_name] = False
        data.attributes[TableAnnotation.gene_id_column] = 'genes'

        self.__initialization(self.human_tax_id)
        self.component.initialize_custom_gene_sets(data)

        self.assertTrue(self.component.gene_sets_combobox.isEnabled())
        self.assertEqual(self.component.custom_gene_set_hierarchy, (data.name,))
        self.assertEqual(self.component._get_selection(), [(data.name,)])
        self.assertIn((data.name,), self.component.gene_sets.map_hierarchy_to_sets().keys())

        emitted_signals: QSignalSpy = QSignalSpy(self.component.selection_changed)
        simulate.combobox_run_through_all(self.component.gene_sets_combobox)
        self.assertEqual(len(emitted_signals), len(data.domain.metas) + len(data.domain.class_vars))

        self.component.initialize_custom_gene_sets(None)
        self.assertNotIn((data.name,), self.component.gene_sets.map_hierarchy_to_sets().keys())
        self.assertIsNone(self.component.custom_gene_set_hierarchy)
        self.assertTrue(self.component.custom_gs_box.isHidden())


if __name__ == "__main__":
    unittest.main()



import random
import unittest
from datetime import datetime, timezone
from unittest.mock import call, patch

from orangewidget.tests.base import GuiTest, WidgetTest

from AnyQt.QtCore import pyqtSignal
from AnyQt.QtTest import QSignalSpy

from Orange.widgets.widget import OWWidget
from Orange.widgets.settings import SettingProvider
from Orange.widgets.credentials import CredentialManager
from Orange.widgets.tests.utils import simulate

from orangecontrib.bioinformatics.widgets.OWGenialisExpressions import (
    DEFAULT_URL,
    SortBy,
    SignInForm,
    ItemsPerPage,
    PaginationComponent,
    FilterByDateModified,
    ResolweAuthException,
    OWGenialisExpressions,
    CollapsibleFilterComponent,
)


class MockWidget(OWWidget):
    name = 'MockWidget'
    want_main_area = False

    filter_component = SettingProvider(CollapsibleFilterComponent)
    pagination_component = SettingProvider(PaginationComponent)

    pagination_availability = pyqtSignal(bool, bool)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filter_component = CollapsibleFilterComponent(self, self.controlArea)
        self.pagination_component = PaginationComponent(self, self.controlArea)
        self.sign_in_dialog = SignInForm(self)


class TestCollapsibleFilterComponent(WidgetTest):
    def setUp(self):
        self.widget = MockWidget()
        self.component = self.widget.filter_component
        self.emitted_signals: QSignalSpy = QSignalSpy(self.component.options_changed)

    def test_filter_by_modified(self):
        # check if correct option is set by default
        self.assertEqual(FilterByDateModified.any_time, self.component.filter_by_modified)
        # run trough all the options
        simulate.combobox_run_through_all(self.component.filter_modified)
        self.assertEqual(FilterByDateModified.past_month, self.component.filter_by_modified)

        # check if signal is emitted on every option changed
        self.assertEqual(len(FilterByDateModified.labels()), len(self.emitted_signals))

    def test_sorting(self):
        # check if correct option is set by default
        self.assertEqual(SortBy.newest_first, self.component.sort_by)
        # run trough all the options
        simulate.combobox_run_through_all(self.component.sorting)
        self.assertEqual(SortBy.oldest_first, self.component.sort_by)

        # check if signal is emitted on every option changed
        self.assertEqual(len(SortBy.labels()), len(self.emitted_signals))

    def test_full_text_search(self):
        search_parameter = 'foobar'
        # Default sorting should be set initially
        self.assertEqual(self.component.sort_by, SortBy.newest_first)
        # simulate user input
        self.component.filter_full_text.setText(search_parameter)
        self.component.on_filter_full_text_changed()
        # check if setting is correctly set
        self.assertEqual(self.component.filter_by_full_text, search_parameter)
        # check if sorting is set accordingly
        self.assertEqual(SortBy.relevance, self.component.sort_by)


class TestPaginationComponent(WidgetTest):
    def setUp(self):
        self.widget = MockWidget()
        self.component = self.widget.pagination_component

    def test_pagination_availability(self):
        self.widget.pagination_availability.emit(False, False)
        self.assertFalse(self.component.page_left_btn.isEnabled())
        self.assertFalse(self.component.page_right_btn.isEnabled())

        self.widget.pagination_availability.emit(True, False)
        self.assertFalse(self.component.page_left_btn.isEnabled())
        self.assertTrue(self.component.page_right_btn.isEnabled())

        self.widget.pagination_availability.emit(False, True)
        self.assertTrue(self.component.page_left_btn.isEnabled())
        self.assertFalse(self.component.page_right_btn.isEnabled())

        self.widget.pagination_availability.emit(True, True)
        self.assertTrue(self.component.page_left_btn.isEnabled())
        self.assertTrue(self.component.page_right_btn.isEnabled())

    def test_page_limit(self):
        # test default option
        self.assertTrue(self.component.page_limit.buttons[ItemsPerPage.min].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.med].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.max].isChecked())
        self.assertEqual(0, self.component.offset)
        self.assertEqual(1, self.component.current_page)

        # change current page
        self.component.right_btn_pressed()
        self.assertEqual(ItemsPerPage.values()[ItemsPerPage.min], self.component.offset)
        self.assertEqual(2, self.component.current_page)

        # change page limit
        self.component.items_per_page = ItemsPerPage.max
        self.component.on_limit_changed()
        self.assertEqual(0, self.component.offset)
        self.assertEqual(1, self.component.current_page)
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.min].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.med].isChecked())
        self.assertTrue(self.component.page_limit.buttons[ItemsPerPage.max].isChecked())

        # change current page
        self.component.right_btn_pressed()
        self.assertEqual(ItemsPerPage.values()[ItemsPerPage.max], self.component.offset)
        self.assertEqual(2, self.component.current_page)

    def test_pagination(self):
        # change page limit
        self.component.items_per_page = ItemsPerPage.med
        self.component.on_limit_changed()
        self.assertEqual(0, self.component.offset)
        self.assertEqual(1, self.component.current_page)
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.min].isChecked())
        self.assertTrue(self.component.page_limit.buttons[ItemsPerPage.med].isChecked())
        self.assertFalse(self.component.page_limit.buttons[ItemsPerPage.max].isChecked())

        current_offset = 0

        for page in [2, 3, 4]:
            self.component.right_btn_pressed()
            current_offset += ItemsPerPage.values()[ItemsPerPage.med]
            self.assertEqual(current_offset, self.component.offset)
            self.assertEqual(page, self.component.current_page)

        for page in [3, 2, 1]:
            self.component.left_btn_pressed()
            current_offset -= ItemsPerPage.values()[ItemsPerPage.med]
            self.assertEqual(current_offset, self.component.offset)
            self.assertEqual(page, self.component.current_page)

        # test left pagination limit
        for _ in range(3):
            self.component.left_btn_pressed()
            self.assertEqual(0, self.component.offset)
            self.assertEqual(1, self.component.current_page)


class TestSignInForm(GuiTest):
    @patch(
        'orangecontrib.bioinformatics.widgets.OWGenialisExpressions.CREDENTIAL_MANAGER_SERVICE',
        'resolwe_credentials_test',
    )
    @patch('orangecontrib.bioinformatics.widgets.OWGenialisExpressions.connect')
    def test_dialog_success(self, mocked_connect):
        widget = MockWidget()
        dialog = SignInForm(widget)
        self.assertFalse(dialog.sign_in_btn.isEnabled())
        dialog.sign_in()
        mocked_connect.assert_not_called()
        self.assertIsNone(dialog.resolwe_instance)

        dialog.username_line_edit.setText('foo')
        dialog.sign_in()
        mocked_connect.assert_not_called()

        dialog.password_line_edit.setText('bar')
        self.assertTrue(dialog.sign_in_btn.isEnabled())
        dialog.sign_in()

        mocked_connect.assert_called_once()
        self.assertTrue(dialog.error_msg.isHidden())
        self.assertIsNotNone(dialog.resolwe_instance)

        # cleanup
        cm = CredentialManager('resolwe_credentials_test')
        del cm.username
        del cm.password

    @patch(
        'orangecontrib.bioinformatics.widgets.OWGenialisExpressions.CREDENTIAL_MANAGER_SERVICE',
        'resolwe_credentials_test',
    )
    @patch('orangecontrib.bioinformatics.widgets.OWGenialisExpressions.connect', side_effect=ResolweAuthException())
    def test_dialog_fail(self, mocked_connect):
        widget = MockWidget()
        dialog = SignInForm(widget)
        dialog.username_line_edit.setText('foo')
        dialog.password_line_edit.setText('bar')

        self.assertTrue(dialog.error_msg.isHidden())
        dialog.sign_in()
        self.assertFalse(dialog.error_msg.isHidden())

        mocked_connect.assert_called_once()

        # cleanup
        cm = CredentialManager('resolwe_credentials_test')
        del cm.username
        del cm.password


def collection_generator(num_of_collections):
    return [
        {
            'id': int(i),
            'name': f'Collection name {i}',
            'contributor': {'first_name': f'Foo_{i}', 'last_name': f'Bar_{i}', 'username': f'FooBar_{i}'},
            'description': f'Test description {1}',
            'data_count': random.randint(0, 10),
            'entity_count': random.randint(0, 10),
            'tags': ['community:expressions'],
            'created': datetime.strftime(datetime.now().replace(tzinfo=timezone.utc), '%Y-%m-%dT%H:%M:%S.%f%z'),
            'modified': datetime.strftime(datetime.now().replace(tzinfo=timezone.utc), '%Y-%m-%dT%H:%M:%S.%f%z'),
        }
        for i in range(num_of_collections)
    ]


class TestOWGenialisExpressions(WidgetTest):
    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    @patch('orangecontrib.bioinformatics.widgets.OWGenialisExpressions.OWGenialisExpressions.sign_in')
    def test_signin_popup(self, mock_sign_in, _):
        self.widget = self.create_widget(OWGenialisExpressions)
        self.widget.sign_in_btn.click()
        mock_sign_in.assert_has_calls([call(silent=True), call(False)])

    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    def test_widget_initialization(self, mock_resolwe_api):
        count = 2
        mock_resolwe_api.return_value.get_collections.return_value = {
            'count': count,
            'next': None,
            'previous': None,
            'results': collection_generator(count),
        }
        self.widget = self.create_widget(OWGenialisExpressions)

        # Check if calls are made with defaults
        offset = 0
        limit = ItemsPerPage.values()[ItemsPerPage.min]
        ordering = SortBy.values()[SortBy.newest_first]
        mock_resolwe_api.assert_called_once_with(None, None, DEFAULT_URL)
        mock_resolwe_api.return_value.get_collections.assert_called_once_with(
            limit=limit, offset=offset, ordering=ordering
        )
        mock_resolwe_api.return_value.get_species.assert_called_once_with(list(range(count)))

        self.assertEqual(count, self.widget.model.rowCount())

    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    def test_sign_out(self, mock_resolwe_api):
        # collections before/after sign out
        count_before = 10
        count_after = 2

        mock_resolwe_api.return_value.get_collections.return_value = {
            'count': count_before,
            'next': None,
            'previous': None,
            'results': collection_generator(count_before),
        }
        self.widget = self.create_widget(OWGenialisExpressions)
        self.assertEqual(count_before, self.widget.model.rowCount())

        # check if state resets after sign out
        mock_resolwe_api.return_value.get_collections.return_value = {
            'count': count_after,
            'next': None,
            'previous': None,
            'results': collection_generator(count_after),
        }
        self.widget.sign_out_btn.click()
        self.assertEqual(count_after, self.widget.model.rowCount())

    @patch('orangecontrib.bioinformatics.resolwe.ResolweAPI', spec=True)
    def test_query_params(self, mock_resolwe_api):
        self.widget = self.create_widget(OWGenialisExpressions)

        full_text = 'test_fts'
        self.widget.filter_component.filter_full_text.setText(full_text)

        col_name = 'test_name'
        self.widget.filter_component.filter_name.setText(col_name)

        col_owner = 'test_owner'
        self.widget.filter_component.filter_owner.setText(col_owner)

        col_contrib = 'test_contrib'
        self.widget.filter_component.filter_contrib.setText(col_contrib)

        self.widget.filter_component.filter_by_modified = FilterByDateModified.past_hour

        self.widget.on_filter_changed()
        _, query_params = mock_resolwe_api.return_value.get_collections.call_args
        self.assertIn('limit', query_params)
        self.assertIn('offset', query_params)
        self.assertIn('ordering', query_params)
        self.assertIn('ordering', query_params)
        self.assertIn('text', query_params)
        self.assertIn('name__icontains', query_params)
        self.assertIn('contributor_name', query_params)
        self.assertIn('owners_name', query_params)
        self.assertIn('modified__gte', query_params)
        self.assertIn(full_text, query_params.values())
        self.assertIn(col_name, query_params.values())
        self.assertIn(col_owner, query_params.values())
        self.assertIn(col_contrib, query_params.values())


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import setuptools

if __name__ == '__main__':
    setuptools.setup(
        use_scm_version=True,
        setup_requires=['setuptools-scm', 'setuptools>=40.0'],
        install_requires=[
            'Orange3>=3.22.0',
            'scipy>=1.5.0',
            'pyclipper>=1.2.0',
            'point-annotator~=2.0',
            'requests',
            'requests-cache',
            'serverfiles',
            'resdk',
            'genesis-pyapi',
            # Versions are determined by Orange
            'numpy',
        ],
        extras_require={
            'doc': ['sphinx', 'recommonmark'],
            'test': [
                'flake8',
                'flake8-comprehensions',
                'flake8-black',
                'pep8-naming',
                'isort',
                'pre-commit',
                'pytest',
                'coverage',
                'codecov',
            ],
        },
    )



""" Bioinformatics add-on for Orange3 """

from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution('orange3-bioinformatics').version
except DistributionNotFound:
    # package is not installed
    pass



import unittest
from unittest.mock import MagicMock

from resdk.resources.sample import Sample

from Orange.data import TimeVariable, StringVariable, DiscreteVariable, ContinuousVariable

from orangecontrib.bioinformatics.resolwe.sample_metadata import (
    descriptors,
    flatten_descriptor,
    flatten_descriptor_schema,
)

SCHEMA = [
    {
        'name': 'section_1',
        'label': 'Section 1',
        'group': [
            {'name': 'field_1', 'type': 'basic:string:', 'label': 'Field 1'},
            {
                'name': 'field_2',
                'type': 'basic:string:',
                'label': 'Field 2',
                'choices': [
                    {'label': 'Choice 1', 'value': 'choice_1'},
                    {'label': 'Choice 2', 'value': 'choice_2'},
                    {'label': 'Choice 3', 'value': 'choice_3'},
                    {'label': 'Choice 4', 'value': 'choice_4'},
                ],
            },
        ],
    },
    {
        'name': 'section_2',
        'label': 'Section 2',
        'group': [
            {
                'name': 'hidden_section_3',
                'group': [
                    {'name': 'field_3', 'type': 'basic:decimal:', 'label': 'Field 3'},
                    {'name': 'field_4', 'type': 'basic:integer:', 'label': 'Field 4'},
                ],
            },
            {'name': 'field_5', 'type': 'basic:boolean:', 'label': 'Field 5'},
            {'name': 'field_6', 'type': 'basic:date:', 'label': 'Field 6'},
            {'name': 'field_7', 'type': 'basic:datetime:', 'label': 'Field 7'},
            {'name': 'field_1', 'type': 'basic:string:', 'label': 'Field 1'},
        ],
    },
]

DESCRIPTOR_SCHEMA = {'id': 1, 'slug': 'test-schema', 'name': 'Test schema', 'version': '1.0.0', 'schema': SCHEMA}


class TestSampleMetadata(unittest.TestCase):
    def test_flatten_descriptor_schema(self):
        sample = Sample(id=1, descriptor_schema=DESCRIPTOR_SCHEMA, resolwe=MagicMock())
        schema, _ = flatten_descriptor_schema(sample.descriptor_schema.schema)

        expected_keys = [
            'section_1.field_1',
            'section_1.field_2',
            'hidden_section_3.field_3',
            'hidden_section_3.field_4',
            'section_2.field_5',
            'section_2.field_6',
            'section_2.field_7',
            'section_2.field_1',
        ]
        self.assertTrue(len(schema.keys()) == 8)
        self.assertTrue(all(key in expected_keys for key in schema.keys()))

        # test if fields with multiple choices are properly handled
        discrete_field = schema['section_1.field_2']
        self.assertTrue(len(discrete_field.keys()) == 4)
        self.assertEqual(
            ('choice_1', 'choice_2', 'choice_3', 'choice_4'),
            tuple(choice['value'] for choice in discrete_field['choices']),
        )

    def test_flatten_descriptor(self):
        descriptor = {
            'section_1': {'field_1': 'field_1_value'},
            'section_2': {'hidden_section_3': {'field_3': 'field_3_value'}, 'field_5': 'field_5_value'},
        }
        sample = Sample(id=1, descriptor=descriptor, resolwe=MagicMock())
        schema = flatten_descriptor(sample.descriptor)

        expected_keys = ['section_1.field_1', 'hidden_section_3.field_3', 'section_2.field_5']
        self.assertTrue(len(schema.keys()) == 3)
        self.assertTrue(all(key in expected_keys for key in schema.keys()))

    def test_descriptors(self):
        samples = []
        descriptor = {
            'section_1': {'field_1': 'field_1_value1', 'field_2': 'choice_1'},
            'section_2': {
                'hidden_section_3': {'field_3': 123.456, 'field_4': 123},
                'field_5': True,
                'field_6': '2020-11-01',
                'field_7': '2020-11-02 12:00:00',
                'field_1': 'test_dup_field1',
            },
        }
        samples.append(
            Sample(
                name='sample1',
                slug='sample2',
                id=1,
                descriptor=descriptor,
                descriptor_schema=DESCRIPTOR_SCHEMA,
                resolwe=MagicMock(),
            )
        )

        descriptor = {
            'section_1': {'field_1': 'field_1_value2', 'field_2': 'choice_2'},
            'section_2': {
                'hidden_section_3': {'field_3': 123.456, 'field_4': 123},
                'field_5': False,
                'field_6': '2020-11-01',
                'field_7': '2020-11-02 13:00:00',
                'field_1': 'test_dup_field2',
            },
        }
        samples.append(
            Sample(
                name='sample2',
                slug='sample2',
                id=2,
                descriptor=descriptor,
                descriptor_schema=DESCRIPTOR_SCHEMA,
                resolwe=MagicMock(),
            )
        )

        descriptor = {
            'section_1': {'field_1': 'field_1_value3', 'field_2': 'choice_3'},
            'section_2': {
                'hidden_section_3': {'field_3': 123.456, 'field_4': 123},
                'field_5': False,
                'field_6': '2020-11-01',
                'field_7': '2020-11-02 14:00:00',
                'field_1': 'test_dup_field3',
            },
        }
        samples.append(
            Sample(
                name='sample3',
                slug='sample3',
                id=3,
                descriptor=descriptor,
                descriptor_schema=DESCRIPTOR_SCHEMA,
                resolwe=MagicMock(),
            )
        )

        metadata = descriptors(samples)
        self.assertTrue(len(metadata.keys()) == 11)
        for var in metadata.keys():
            self.assertIsInstance(var, (ContinuousVariable, DiscreteVariable, StringVariable, TimeVariable))

        # test for prefix in duplicate fields
        variable_names = {var.name for var in metadata.keys()}
        self.assertTrue('Field 1' not in variable_names)
        self.assertTrue('Section 1 Field 1' in variable_names)
        self.assertTrue('Section 2 Field 1' in variable_names)

        # ensure that discrete variables match sample descriptor
        self.assertEqual(list(metadata.get(DiscreteVariable('Field 5'))), [1, 0, 0])
        self.assertEqual(list(metadata.get(DiscreteVariable('Field 2'))), [0, 1, 2])



""" Databases update widget """
import os
import sys
import json
import threading
from shutil import copyfile
from datetime import datetime as d_time
from functools import partial
from collections import OrderedDict, namedtuple

from requests.exceptions import Timeout, ConnectionError

from AnyQt.QtCore import Qt, QSize, Signal, QThreadPool
from AnyQt.QtWidgets import (
    QLabel,
    QDialog,
    QWidget,
    QCheckBox,
    QComboBox,
    QLineEdit,
    QFileDialog,
    QHBoxLayout,
    QPushButton,
    QToolButton,
    QTreeWidget,
    QVBoxLayout,
    QApplication,
    QTreeWidgetItem,
    QDialogButtonBox,
    QAbstractItemView,
    QStyledItemDelegate,
)

from Orange.widgets import gui
from Orange.widgets.widget import OWWidget

from orangecontrib.bioinformatics.utils import serverfiles
from orangecontrib.bioinformatics.geneset import DOMAIN as gene_sets_domain
from orangecontrib.bioinformatics.geneset import filename
from orangecontrib.bioinformatics.go.config import DOMAIN as gene_ontology_domain
from orangecontrib.bioinformatics.go.config import FILENAME_ANNOTATION
from orangecontrib.bioinformatics.ncbi.taxonomy import common_taxids, common_taxid_to_name, species_name_to_taxid
from orangecontrib.bioinformatics.widgets.utils.gui import TokenListCompleter
from orangecontrib.bioinformatics.widgets.utils.concurrent import Worker

# File states
AVAILABLE, CURRENT, OUTDATED, DEPRECATED, USER_FILE = range(5)
# File sources
SOURCE_SERVER = 'server_file'  # files on the serverfiles-bio repository
SOURCE_USER = 'user_file'  # user defined files
INFO_FILE_SCHEMA = {
    'domain': None,
    'filename': None,
    'source': None,
    'title': None,
    'tags': [],
    'size': None,
    'datetime': None
    # used only if files are compressed
    # 'uncompressed': None,
    # 'compression': None,
}


def file_size_bytes(file_path):
    """ returns file size in bytes """
    return os.stat(file_path).st_size


def create_info_file(file_path, **kwargs):
    info_dict = OrderedDict(INFO_FILE_SCHEMA)

    info_dict.update(**kwargs)
    info_dict['datetime'] = '{0:%Y-%m-%d %H:%M:%S.%f}'.format(d_time.today())
    info_dict['size'] = file_size_bytes(file_path)

    with open(file_path + '.info', 'wt') as f:
        json.dump(info_dict, f)


def create_folder(path):
    try:
        os.makedirs(path)
    except OSError:
        if os.path.exists(path):
            pass
        else:
            # There was an error on creation, so make sure we know about it
            raise


class UpdateOptionsItemDelegate(QStyledItemDelegate):
    """ An item delegate for the updates tree widget.

    note:
        Must be a child of a QTreeWidget.

    """

    def sizeHint(self, option, index):
        size = QStyledItemDelegate.sizeHint(self, option, index)
        parent = self.parent()
        item = parent.itemFromIndex(index)
        widget = parent.itemWidget(item, 0)
        if widget:
            size = QSize(size.width(), widget.sizeHint().height() / 2)
        return size


file_state = namedtuple('file_state', ['info_local', 'info_server', 'state'])

header_labels = ['', 'Title', 'Update', 'Updated', 'Size', 'Source']
header_index = namedtuple('header_index', ['Download'] + header_labels[1:])
header = header_index(*[index for index, _ in enumerate(header_labels)])


def UpdateItem_match(item, string):
    """
    Return `True` if the `UpdateItem` item contains a string in tags
    or in the title.

    """
    string = string.lower()
    return any(string.lower() in tag.lower() for tag in item.tags + [item.title])


def evaluate_all_info(local, server):
    """ Return FileState

    Args:
        local: info files from LocalFiles
        server: info files from ServerFiles

    """
    files = set(local.keys()).union(server.keys())

    for domain, file_name in sorted(files):
        yield FileState(domain, file_name, server.get((domain, file_name), None), local.get((domain, file_name), None))


def evaluate_files_state(progress_callback):
    progress_callback.emit()
    files = []

    # fetch remote info
    try:
        server_info = serverfiles.ServerFiles().allinfo()
    except (Timeout, ConnectionError) as e:
        raise e
    progress_callback.emit()

    # fetch local info
    local_info = serverfiles.allinfo()

    all_info = set(local_info.keys()).union(server_info.keys())

    for domain, file_name in sorted(all_info):
        files.append(
            FileState(
                domain,
                file_name,
                server_info.get((domain, file_name), None),
                local_info.get((domain, file_name), None),
            )
        )
    progress_callback.emit()
    return files


def download_server_file(fs, index, progress_callback):
    try:
        serverfiles.download(fs.domain, fs.filename, callback=progress_callback.emit)
    except Exception:
        # send FileState and index with Exception
        raise ValueError(fs, index)

    return fs, index


class OWDatabasesUpdate(OWWidget):

    name = "Databases Update"
    description = "Update local systems biology databases."
    icon = "../widgets/icons/OWDatabasesUpdate.svg"
    priority = 1

    inputs = []
    outputs = []

    want_main_area = False

    def __init__(self, parent=None, signalManager=None, name="Databases update"):
        OWWidget.__init__(self, parent, signalManager, name, wantMainArea=False)

        self.searchString = ""

        fbox = gui.widgetBox(self.controlArea, "Filter")
        self.completer = TokenListCompleter(self, caseSensitivity=Qt.CaseInsensitive)
        self.lineEditFilter = QLineEdit(textChanged=self.search_update)
        self.lineEditFilter.setCompleter(self.completer)

        fbox.layout().addWidget(self.lineEditFilter)

        box = gui.widgetBox(self.controlArea, "Files")
        self.filesView = QTreeWidget(self)
        self.filesView.setHeaderLabels(header_labels)
        self.filesView.setRootIsDecorated(False)
        self.filesView.setUniformRowHeights(True)
        self.filesView.setSelectionMode(QAbstractItemView.NoSelection)
        self.filesView.setSortingEnabled(True)
        self.filesView.sortItems(header.Title, Qt.AscendingOrder)
        self.filesView.setItemDelegateForColumn(0, UpdateOptionsItemDelegate(self.filesView))

        self.filesView.model().layoutChanged.connect(self.search_update)

        box.layout().addWidget(self.filesView)

        layout = QHBoxLayout()
        gui.widgetBox(self.controlArea, margin=0, orientation=layout)

        self.updateButton = gui.button(
            box, self, "Update all", callback=self.update_all, tooltip="Update all updatable files"
        )

        self.downloadButton = gui.button(
            box, self, "Download all", callback=self.download_filtered, tooltip="Download all filtered files shown"
        )

        self.cancelButton = gui.button(
            box, self, "Cancel", callback=self.cancel_active_threads, tooltip="Cancel scheduled downloads/updates."
        )

        self.addButton = gui.button(
            box, self, "Add ...", callback=self.__handle_dialog, tooltip="Add files for personal use."
        )

        layout.addWidget(self.updateButton)
        layout.addWidget(self.downloadButton)
        layout.addWidget(self.cancelButton)
        layout.addStretch()
        layout.addWidget(self.addButton)

        # Enable retryButton once connection is established
        # self.retryButton = gui.button(
        #     box, self, "Reconnect", callback=self.initialize_files_view
        # )
        # self.retryButton.hide()

        self.resize(800, 600)

        self.update_items = []
        self._dialog = None
        self.progress_bar = None

        # threads
        self.threadpool = QThreadPool(self)
        # self.threadpool.setMaxThreadCount(1)
        self.workers = []

        self.initialize_files_view()

    def __handle_dialog(self):
        if not self._dialog:
            self._dialog = FileUploadHelper(self)
        self._dialog.show()

    def __progress_advance(self):
        # GUI should be updated in main thread. That's why we are calling advance method here
        if self.progress_bar:
            self.progress_bar.advance()

    def handle_worker_exception(self, ex):
        self.progress_bar.finish()
        self.setStatusMessage('')

        if isinstance(ex, ConnectionError):
            # TODO: set warning messages
            pass

        print(ex)

    def initialize_files_view(self):
        # self.retryButton.hide()

        # clear view
        self.filesView.clear()
        # init progress bar
        self.progress_bar = gui.ProgressBar(self, iterations=3)
        # status message
        self.setStatusMessage('initializing')

        worker = Worker(evaluate_files_state, progress_callback=True)
        worker.signals.progress.connect(self.__progress_advance)
        worker.signals.result.connect(self.set_files_list)
        worker.signals.error.connect(self.handle_worker_exception)

        # move download process to worker thread
        self.threadpool.start(worker)
        self.setEnabled(False)

    def __create_action_button(self, fs, retry=None):
        if not fs.state not in [OUTDATED, USER_FILE] or not retry:
            self.filesView.setItemWidget(fs.tree_item, header.Update, None)

        button = QToolButton(None)
        if not retry:
            if fs.state == OUTDATED:
                button.setText('Update')
                button.clicked.connect(partial(self.submit_download_task, fs.domain, fs.filename, True))
            elif fs.state == USER_FILE:
                if not fs.info_server:
                    button.setText('Remove')
                    button.clicked.connect(partial(self.submit_remove_task, fs.domain, fs.filename))
                else:
                    button.setText('Use server version')
                    button.clicked.connect(partial(self.submit_download_task, fs.domain, fs.filename, True))
        else:
            button.setText('Retry')
            button.clicked.connect(partial(self.submit_download_task, fs.domain, fs.filename, True))

        button.setMaximumWidth(120)
        button.setMaximumHeight(20)
        button.setMinimumHeight(20)

        if sys.platform == "darwin":
            button.setAttribute(Qt.WA_MacSmallSize)

        self.filesView.setItemWidget(fs.tree_item, header.Update, button)

    def set_files_list(self, result):
        """ Set the files to show.
        """
        assert threading.current_thread() == threading.main_thread()
        self.progress_bar.finish()
        self.setStatusMessage('')
        self.setEnabled(True)

        self.update_items = result
        all_tags = set()

        for fs in self.update_items:
            fs.tree_item = FileStateItem(fs)
            fs.download_option = DownloadOption(state=fs.state)

            fs.download_option.download_clicked.connect(partial(self.submit_download_task, fs.domain, fs.filename))
            fs.download_option.remove_clicked.connect(partial(self.submit_remove_task, fs.domain, fs.filename))

        # add widget items to the QTreeWidget
        self.filesView.addTopLevelItems([fs.tree_item for fs in self.update_items])

        # add action widgets to tree items
        for fs in self.update_items:
            self.filesView.setItemWidget(fs.tree_item, header.Download, fs.download_option)
            if fs.state in [USER_FILE, OUTDATED]:
                self.__create_action_button(fs)

            all_tags.update(fs.tags)

        self.filesView.setColumnWidth(header.Download, self.filesView.sizeHintForColumn(header.Download))

        for column in range(1, len(header_labels)):
            self.filesView.resizeColumnToContents(column)

        hints = [hint for hint in sorted(all_tags) if not hint.startswith("#")]
        self.completer.setTokenList(hints)
        self.search_update()
        self.toggle_action_buttons()
        self.cancelButton.setEnabled(False)

    def toggle_action_buttons(self):
        selected_items = [fs for fs in self.update_items if not fs.tree_item.isHidden()]

        def button_check(sel_items, state, button):
            for item in sel_items:
                if item.state != state:
                    button.setEnabled(False)
                else:
                    button.setEnabled(True)
                    break

        button_check(selected_items, OUTDATED, self.updateButton)
        button_check(selected_items, AVAILABLE, self.downloadButton)

    def search_update(self, searchString=None):
        strings = str(self.lineEditFilter.text()).split()
        for fs in self.update_items:
            hide = not all(UpdateItem_match(fs, string) for string in strings)
            fs.tree_item.setHidden(hide)
        self.toggle_action_buttons()

    def update_all(self):
        for fs in self.update_items:
            if fs.state == OUTDATED and not fs.tree_item.isHidden():
                self.submit_download_task(fs.domain, fs.filename)

    def download_filtered(self):
        for fs in self.update_items:
            if not fs.tree_item.isHidden() and fs.state in [AVAILABLE, OUTDATED]:
                self.submit_download_task(fs.domain, fs.filename, start=False)

        self.run_download_tasks()

    def submit_download_task(self, domain, filename, start=True):
        """ Submit the (domain, filename) to be downloaded/updated.
        """
        # get selected tree item
        index = self.tree_item_index(domain, filename)
        fs = self.update_items[index]

        worker = Worker(download_server_file, fs, index, progress_callback=True)
        worker.signals.progress.connect(self.__progress_advance)
        worker.signals.result.connect(self.on_download_finished)
        worker.signals.error.connect(self.on_download_exception)

        self.workers.append(worker)

        if start:
            self.run_download_tasks()

    def run_download_tasks(self):
        self.cancelButton.setEnabled(True)
        # init progress bar

        self.progress_bar = gui.ProgressBar(self, iterations=len(self.workers) * 100)

        # status message
        self.setStatusMessage('downloading')

        # move workers to threadpool
        [self.threadpool.start(worker) for worker in self.workers]
        self.filesView.setDisabled(True)
        # reset list of workers
        self.workers = []

    def on_download_exception(self, ex):
        assert threading.current_thread() == threading.main_thread()
        self.progress_bar.finish()
        self.setStatusMessage('')
        print(ex)
        if isinstance(ex, ValueError):
            fs, index = ex.args

            # restore state and retry
            fs.refresh_state()
            fs.tree_item.update_data(fs)
            fs.download_option.state = fs.state
            self.__create_action_button(fs, retry=True)

    def on_download_finished(self, result):
        assert threading.current_thread() == threading.main_thread()

        # We check if all workers have completed. If not, continue
        if self.progress_bar.count == 100 or self.threadpool.activeThreadCount() == 0:
            self.filesView.setDisabled(False)
            self.progress_bar.finish()
            self.setStatusMessage('')

        fs, index = result
        # re-evaluate File State
        info = serverfiles.info(fs.domain, fs.filename)
        fs.refresh_state(info_local=info, info_server=info)
        # reinitialize treeWidgetItem
        fs.tree_item.update_data(fs)
        # reinitialize OptionWidget
        fs.download_option.state = fs.state
        self.filesView.setItemWidget(fs.tree_item, header.Update, None)

        self.toggle_action_buttons()
        for column in range(1, len(header_labels)):
            self.filesView.resizeColumnToContents(column)

    def submit_remove_task(self, domain, filename):
        serverfiles.LOCALFILES.remove(domain, filename)

        index = self.tree_item_index(domain, filename)
        fs = self.update_items[index]

        if fs.state == USER_FILE:
            self.filesView.takeTopLevelItem(self.filesView.indexOfTopLevelItem(fs.tree_item))
            self.update_items.remove(fs)
            # self.filesView.removeItemWidget(index)
        else:
            # refresh item state
            fs.info_local = None
            fs.refresh_state()
            # reinitialize treeWidgetItem
            fs.tree_item.update_data(fs)
            # reinitialize OptionWidget
            fs.download_option.state = fs.state

        self.toggle_action_buttons()

    def cancel_active_threads(self):
        """ Cancel all pending update/download tasks (that have not yet started).
        """
        if self.threadpool:
            self.threadpool.clear()

    def tree_item_index(self, domain, filename):
        for i, fs in enumerate(self.update_items):
            if fs.domain == domain and fs.filename == filename:
                return i
        raise ValueError("%r, %r not in update list" % (domain, filename))

    def onDeleteWidget(self):
        self.cancel_active_threads()
        OWWidget.onDeleteWidget(self)


class DownloadOption(QWidget):
    """ A Widget with download/update/remove options.
    """

    download_clicked = Signal()
    remove_clicked = Signal()

    def __init__(self, state=AVAILABLE, parent=None):
        QWidget.__init__(self, parent)
        layout = QHBoxLayout()
        layout.setSpacing(1)
        layout.setContentsMargins(1, 1, 1, 1)

        self.checkButton = QCheckBox()

        layout.addWidget(self.checkButton)
        self.setLayout(layout)

        self.setMinimumHeight(20)
        self.setMaximumHeight(20)

        self._state = state
        self._update()

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        self._state = state
        self._update()

    def _update(self):
        self.checkButton.setDisabled(False)

        if self.state == AVAILABLE:
            self.checkButton.setChecked(False)
        elif self.state == CURRENT:
            self.checkButton.setChecked(True)
        elif self.state == OUTDATED:
            self.checkButton.setChecked(True)
        elif self.state == DEPRECATED:
            self.checkButton.setChecked(True)
        elif self.state == USER_FILE:
            self.checkButton.setChecked(False)
            self.checkButton.setDisabled(True)
        else:
            raise ValueError("Invalid state %r" % self.state)

        try:
            self.checkButton.clicked.disconnect()  # Remove old signals if they exist
        except Exception:
            pass

        if not self.checkButton.isChecked():  # Switch signals if the file is present or not
            self.checkButton.clicked.connect(self.download_clicked)
        else:
            self.checkButton.clicked.connect(self.remove_clicked)


class FileState:
    def __init__(self, domain, file_name, info_server, info_local):
        self.domain = domain
        self.filename = file_name

        self.info_server = self.parse_info_file(info_server)
        self.info_local = self.parse_info_file(info_local)

        # self.source = None
        self.state = self.__item_state()

        self.tree_item = None
        self.download_option = None

    def refresh_state(self, info_server=None, info_local=None):
        if info_local:
            self.info_local = self.parse_info_file(info_local)
        if info_server:
            self.info_server = self.parse_info_file(info_server)

        self.state = self.__item_state()

    @property
    def tags(self):
        if self.state in [AVAILABLE]:
            return self.info_server.tags
        elif self.state in [OUTDATED, USER_FILE, CURRENT]:
            return self.info_local.tags
        else:
            return None

    @property
    def title(self):
        if self.state in [AVAILABLE]:
            return self.info_server.title
        elif self.state in [OUTDATED, USER_FILE, CURRENT]:
            return self.info_local.title
        else:
            return None

    @property
    def size(self):
        if self.state in [AVAILABLE]:
            return self.info_server.size
        elif self.state in [OUTDATED, USER_FILE, CURRENT]:
            return self.info_local.size
        else:
            return None

    @property
    def datetime(self):
        if self.state in [USER_FILE, CURRENT, OUTDATED]:
            return self.info_local.datetime
        else:
            return self.info_server.datetime

    @property
    def source(self):
        if self.state is USER_FILE:
            return self.info_local.source.capitalize().replace('_', ' ')
        else:
            return self.info_server.source.capitalize().replace('_', ' ')

    def __item_state(self):
        """ Return item state

        Note:
            available  -> available for download
            current    -> latest version downloaded
            outdated   -> needs update (newer version on serverfiles-bio repository)
            deprecated -> removed from serverfiles-bio repository
            user_file  -> not in serverfiles-bio repository (user defined)

        """

        if not self.info_server and self.info_local:
            # we check source of the file
            if self.info_local.source == SOURCE_USER:
                # this is user defined file
                return USER_FILE

            elif self.info_local.source == SOURCE_SERVER:
                # there is no record of this file on the server
                return DEPRECATED

        if not self.info_local and self.info_server:
            return AVAILABLE

        if self.info_server and self.info_local:
            if not self.info_local.source == SOURCE_USER:

                if self.info_local.datetime < self.info_server.datetime:
                    return OUTDATED
                else:
                    return CURRENT

            else:
                return USER_FILE

    @staticmethod
    def parse_info_file(info):
        """ Parse .info file from JSON like format to namedtuple
        """
        if info is not None:
            if not isinstance(info['datetime'], d_time):
                info['datetime'] = d_time.strptime(info['datetime'], "%Y-%m-%d %H:%M:%S.%f")
            return namedtuple('file_info', info.keys())(**info)


class FileStateItem(QTreeWidgetItem):

    STATE_STRINGS = {
        0: 'not downloaded',
        1: 'downloaded, current',
        2: 'downloaded, needs update',
        3: 'obsolete',
        4: 'custom file',
    }

    #: A role for the state item data.
    StateRole = next(gui.OrangeUserRole)

    # QTreeWidgetItem stores the DisplayRole and EditRole as the same role,
    # so we can't use EditRole to store the actual item data, instead we use
    # custom role.

    #: A custom edit role for the item's data
    #: (QTreeWidget treats Qt.EditRole as a alias for Qt.DisplayRole)
    EditRole2 = next(gui.OrangeUserRole)

    def __init__(self, fs):
        """ A QTreeWidgetItem for displaying a FileState.
        """
        QTreeWidgetItem.__init__(self, type=QTreeWidgetItem.UserType)
        self.update_data(fs)
        self._update_tool_tip(fs)

    def update_data(self, fs):
        self.setData(header.Download, FileStateItem.StateRole, fs.state)

        self.setData(header.Source, Qt.DisplayRole, fs.source)
        self.setData(header.Source, self.EditRole2, fs.source)

        self.setData(header.Title, Qt.DisplayRole, fs.title)
        self.setData(header.Title, self.EditRole2, fs.title)

        if not fs.state == AVAILABLE:
            self.setData(header.Updated, Qt.DisplayRole, fs.datetime.date().isoformat())
            self.setData(header.Updated, self.EditRole2, fs.datetime)
        else:
            self.setData(header.Updated, Qt.DisplayRole, '')
            self.setData(header.Updated, self.EditRole2, '')

        self.setData(header.Size, Qt.DisplayRole, serverfiles.sizeformat(fs.size))
        self.setData(header.Size, self.EditRole2, fs.size)

    def _update_tool_tip(self, fs):
        state_str = self.STATE_STRINGS[fs.state]
        if fs == DEPRECATED:
            diff_date = fs.info_server.datetime - fs.info_local.datetime
        else:
            diff_date = None

        tooltip = "State: {}\nTags: {}".format(state_str, ', '.join(tag for tag in fs.tags if not tag.startswith("#")))

        if fs.state in [CURRENT, OUTDATED, DEPRECATED]:
            tooltip += "\nFile: {}".format(serverfiles.localpath(fs.domain, fs.filename))

        if fs.state == OUTDATED and diff_date:
            tooltip += "\nServer version: {}\nStatus: old {} days".format(fs.datetime, diff_date.days)
        else:
            tooltip += "\nServer version: {}".format(fs.datetime)

        for i in range(1, len(header_labels) - 1):
            self.setToolTip(i, tooltip)

    def __lt__(self, other):
        widget = self.treeWidget()
        column = widget.sortColumn()
        if column == 0:
            role = FileStateItem.StateRole
        else:
            role = self.EditRole2

        left = self.data(column, role)
        right = other.data(column, role)
        try:
            return left < right
        except TypeError:
            pass
        # order lexically by str representation, but ensure `None`
        # always orders on one side
        left = (0, "") if left is None else (1, str(left))
        right = (0, "") if right is None else (1, str(right))
        return left < right


class FileUploadHelper(QDialog):

    # settings
    kegg_domain = 'KEGG'

    supported_domains = OrderedDict({'Gene Ontology': gene_ontology_domain, 'Gene Sets': gene_sets_domain})

    supported_organisms = [common_taxid_to_name(tax_id) for tax_id in common_taxids()]

    hierarchies = {
        'GO - Biological Process': ('GO', 'biological_process'),
        'GO - Molecular Function': ('GO', 'molecular_function'),
        'GO - Cellular Component': ('GO', 'cellular_component'),
        'KEGG - Pathways': ('KEGG', 'pathways'),
        'KEGG - Orthologs': ('KEGG', 'orthologs'),
    }

    def __init__(self, parent=None):
        super(FileUploadHelper, self).__init__(
            parent,
            Qt.Window
            | Qt.WindowTitleHint
            | Qt.CustomizeWindowHint
            | Qt.WindowCloseButtonHint
            | Qt.WindowMaximizeButtonHint,
        )
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowTitle('Add new file')

        self.info_state = INFO_FILE_SCHEMA
        self.layout = QVBoxLayout(self)

        # domain selection combobox
        self.domain_selection = QComboBox()
        self.domain_selection.addItems(self.supported_domains.keys())
        self.domain_selection.currentIndexChanged.connect(self.__on_domain_selection)
        self.__create_selection_row('Domain: ', self.domain_selection)

        # domain selection combobox
        self.hierarchy_selection = QComboBox()
        self.hierarchy_selection.addItems(self.hierarchies.keys())
        self.layout.addWidget(self.hierarchy_selection, alignment=Qt.AlignVCenter)
        self.__on_domain_selection()

        # select organism
        self.organism_selection = QComboBox()
        self.organism_selection.addItems(self.supported_organisms)
        self.__create_selection_row('Organism: ', self.organism_selection)

        # title
        self.line_edit_title = QLineEdit()
        self.__create_selection_row('Title: ', self.line_edit_title)

        # tags
        self.line_edit_tags = QLineEdit()
        self.__create_selection_row('Tags (comma-separated): ', self.line_edit_tags)

        # file selector
        self.file_info = QLabel()
        self.file_select_btn = QPushButton('Select File', self)
        self.file_select_btn.clicked.connect(self.__handle_file_selector)
        self.__create_selection_row(' ', self.file_select_btn)

        # add file info section
        self.layout.addWidget(self.file_info, alignment=Qt.AlignCenter)

        self.layout.addStretch(1)

        # Ok and Cancel buttons
        self.buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)
        self.layout.addWidget(self.buttons, alignment=Qt.AlignJustify)

        self.buttons.accepted.connect(self.__accept)
        self.buttons.rejected.connect(self.__close)

        # path to a selected file
        self.file_path = None

    def __on_domain_selection(self):
        selected = self.__get_selected_domain() == gene_sets_domain
        self.hierarchy_selection.setVisible(selected)

    def __get_selected_domain(self):
        domain_label = list(self.supported_domains.keys())[self.domain_selection.currentIndex()]
        return self.supported_domains[domain_label]

    def __get_selected_hier(self):
        hier_label = list(self.hierarchies.keys())[self.hierarchy_selection.currentIndex()]
        return self.hierarchies[hier_label]

    def __create_selection_row(self, label, widget):
        self.layout.addWidget(QLabel(label), alignment=Qt.AlignLeft)
        self.layout.addWidget(widget, alignment=Qt.AlignVCenter)

    def __accept(self):
        if self.file_path:
            self.info_state = self.__parse_selection()
            self.__move_to_serverfiles_folder(self.file_path)

            self.parent().initialize_files_view()
            self.close()

    def __close(self):
        self.close()

    def closeEvent(self, event):
        # clean-up
        self.parent()._dialog = None

    def __filename(self, domain, organism):
        """ Create filename based od domain name and organism.
        """

        if domain in self.supported_domains.values() and domain == gene_ontology_domain and organism:
            return FILENAME_ANNOTATION.format(organism)

        elif domain in self.supported_domains.values() and domain == gene_sets_domain and organism:
            return filename((self.__get_selected_hier()), organism)

    def __parse_selection(self):
        try:
            domain = self.__get_selected_domain()
            organism = species_name_to_taxid(self.supported_organisms[self.organism_selection.currentIndex()])
        except KeyError as e:
            raise e

        return {
            'domain': domain,
            'organism': organism,
            'filename': self.__filename(domain, organism),
            'title': self.line_edit_title.text(),
            'tags': self.line_edit_tags.text().split(','),
            'source': SOURCE_USER,
        }

    def __move_to_serverfiles_folder(self, selected_file_path):
        domain_path = serverfiles.localpath(self.info_state['domain'])
        file_path = os.path.join(domain_path, self.info_state['filename'])
        create_folder(domain_path)

        try:
            copyfile(selected_file_path, file_path)
        except IOError as e:
            # TODO: handle error properly
            raise e

        # if copy successful create .info file
        create_info_file(file_path, **self.info_state)

    def __handle_file_selector(self):
        self.file_path = QFileDialog.getOpenFileName(self, 'Open File')[0]
        self.file_info.setText('Selected File: {}'.format(os.path.basename(self.file_path)))


if __name__ == "__main__":

    def main_test():
        app = QApplication(sys.argv)
        w = OWDatabasesUpdate()
        w.show()
        w.raise_()
        return app.exec_()

    sys.exit(main_test())
